import React, { Fragment, useState, useEffect } from 'react';
import * as PropTypes from 'prop-types';
import {

    // DataTypeProvider, // used to populate custom filter operation drop down items
    FilteringState,
    SelectionState,
    VirtualTableState,
} from '@devexpress/dx-react-grid';
import {
    DragDropProvider,
    Grid,
    TableColumnReordering,
    TableColumnResizing,
    TableFilterRow,
    TableFixedColumns,
    TableHeaderRow,
    TableSelection,
    VirtualTable,
} from '@devexpress/dx-react-grid-material-ui';

import ActivityIndicator from '../ActivityIndicator/ActivityIndicator';
import IconButton, { ICON_SIZE } from '../IconButton/IconButton';
import TextLabel from '../TextLabel/TextLabel';

import getStyles from './DataTable.styles';

const DataTableCell = VirtualTable.Cell;
const DEFAULT_COLUMN_WIDTH = 200;

/**
 * Enum for filter operations
 * @typedef {Object} FilterOperations
 * @enum {String}
 */
const FILTER_OPERATIONS = {
    CONTAINS: 'contains',
    NOT_CONTAINS: 'notContains',
    STARTS_WITH: 'startsWith',
    ENDS_WITH: 'endsWith',
    EQUAL: 'equal',
    NOT_EQUAL: 'notEqual',
    GREATER_THAN: 'greaterThan',
    GREATER_THAN_OR_EQUAL: 'greaterThanOrEqual',
    LESS_THAN: 'lessThan',
    LESS_THAN_OR_EQUAL: 'lessThanOrEqual',
};

const HeaderCellContent = (props) => {
    console.log('HEADER CELL: RENDER: ', props); // eslint-disable-line
    const {
        column,
        fixedColumns,
        onColumnFrozen,
        onFilterToggle,
        showFilterRow,
    } = props;
    const {
        name,
        title,
    } = column;
    const fixedColumn = fixedColumns[name];
    let fixedIcon = 'lock_open';
    let fixedTitle = 'lock column';
    if (fixedColumn !== undefined) {
        fixedIcon = 'lock';
        fixedTitle = 'unlock column';
    }
    return (
        <Fragment>
            <TextLabel>
                { title }
            </TextLabel>
            <IconButton
                iconName="filter_list"
                iconSize={ ICON_SIZE.SMALL }
                onClick={ () => onFilterToggle(name) }
                title={ `${ showFilterRow ? 'close' : 'open' } filter` } />
            <IconButton
                iconName={ fixedIcon }
                iconSize={ ICON_SIZE.SMALL }
                onClick={ () => onColumnFrozen(name) }
                title={ fixedTitle } />
        </Fragment>
    );
};
HeaderCellContent.propTypes = {
    // Unused available props
    // align: PropTypes.bool,
    // children: PropTypes.oneOfType([
    //     PropTypes.array,
    //     PropTypes.node,
    //     PropTypes.element,
    // ]),

    column: PropTypes.object,
    fixedColumns: PropTypes.object,
    onColumnFrozen: PropTypes.func,
    onFilterToggle: PropTypes.func,
    showFilterRow: PropTypes.bool,
};

const FilterCell = (props) => {
    console.log('FILTER CELL: RENDER: ', props); // eslint-disable-line
    const {
        children,
        column,
        onFilterCancel,
        onFilterReset,
        onFilterSubmit,
        onSortAscending,
        onSortClear,
        onSortDescending,
        sortDirectives,
        ...rest
    } = props;
    const {
        name,
    } = column;
    const sortDirectiveIndex = sortDirectives.findIndex((sortedColumn) => (sortedColumn.columnName === name));
    let sortDirective = {};
    let sortOrder = 0;
    if (sortDirectiveIndex >= 0) {
        sortDirective = sortDirectives[sortDirectiveIndex];
        sortOrder = sortDirectiveIndex + 1;
    }
    const {
        direction,
    } = sortDirective;
    return (
        <TableFilterRow.Cell
            { ...rest }
            column={ column }>
            <div style={ {
                display: 'flex',
                flexDirection: 'column',
            } }>
                <div style={ {
                    display: 'flex',
                    flexDirection: 'row',
                } }>
                    <IconButton
                        badgeText={ direction === 'asc' && sortOrder ? sortOrder.toString() : '' }
                        contextStyles={ {
                            icon: {
                                // Invert sort descending icon for sort ascending as one was not in the core library
                                transform: 'scaleY(-1)',
                            },
                        } }
                        disabled={ direction === 'asc' }
                        iconName="sort"
                        iconSize={ ICON_SIZE.SMALL }
                        onClick={ () => { onSortAscending(name); } }
                        title="sort ascending" />
                    <IconButton
                        badgeText={ direction === 'desc' && sortOrder ? sortOrder.toString() : '' }
                        disabled={ direction === 'desc' }
                        iconName="sort"
                        iconSize={ ICON_SIZE.SMALL }
                        onClick={ () => { onSortDescending(name); } }
                        title="sort descending" />
                    <IconButton
                        disabled={ !direction }
                        iconName="block"
                        iconSize={ ICON_SIZE.SMALL }
                        onClick={ () => { onSortClear(name); } }
                        title="clear sort" />
                </div>
                <div style={ {
                    display: 'flex',
                    flexDirection: 'row',
                } }>
                    { children }
                </div>
                <div style={ {
                    display: 'flex',
                    flexDirection: 'row',
                } }>
                    <IconButton
                        iconName="done"
                        iconSize={ ICON_SIZE.SMALL }
                        onClick={ () => { onFilterSubmit(name); } }
                        title="apply filters" />
                    <IconButton
                        iconName="undo"
                        iconSize={ ICON_SIZE.SMALL }
                        onClick={ () => { onFilterReset(name); } }
                        title="reset filters" />
                    <IconButton
                        iconName="clear"
                        iconSize={ ICON_SIZE.SMALL }
                        onClick={ () => { onFilterCancel(name); } }
                        title="cancel" />
                </div>
            </div>
        </TableFilterRow.Cell>
    );
};
FilterCell.propTypes = {
    // Unused available props
    // colSpan: PropTypes.number,
    // filter: PropTypes.object,
    // filteringEnabled: PropTypes.bool,
    // getMessage: PropTypes.func,
    // onFilter: PropTypes.func,
    // rowSpan: PropTypes.number,
    // tableRow: PropTypes.object,
    // tableColumn: PropTypes.object,

    children: PropTypes.oneOfType([
        PropTypes.array,
        PropTypes.node,
        PropTypes.element,
    ]),
    column: PropTypes.object,
    onFilterCancel: PropTypes.func,
    onFilterReset: PropTypes.func,
    onFilterSubmit: PropTypes.func,
    onSortAscending: PropTypes.func,
    onSortClear: PropTypes.func,
    onSortDescending: PropTypes.func,
    sortDirectives: PropTypes.array, // [ { columnName: String, direction: 'asc'|'desc'|null } ]
};

/**
 * An extensive table grid control
 * @see https://devexpress.github.io/devextreme-reactive/react/grid/
 */
const DataTable = (props) => {
    const {
        cellComponent, // Function
        columns, // [ { name: String, title?: String } ]
        contextStyles, // Object
        defaultColumnWidth, // Number
        filterDirectives, // [ { columnName: String, operation: String, value: any } ]
        filterOperations, // [ { columnName: String, operators: Array<String> } ]
        fixedColumnIndexes, // [ Number ]
        infiniteScrolling, // Boolean
        loading, // Boolean
        onLoadMoreData, // Function { Number } skip, { Number } take
        onColumnFrozen, // Function { String } columnName
        onColumnsReordered, // Function [ String ] array of column names in order
        onColumnsResized, // Function [ { columnName: String, width: Number } ]
        onFilterCancel, // Function { String } columnName
        onFilterChange, // Function [ { columnName: String, operation: String, value: String } ]
        onFilterReset, // Function { String } columnName
        onFilterSubmit, // Function { String } columnName
        onFilterToggle, // Function { String } columnName
        onRowSelectionChange, // Function [ Number ] selected row indexes
        onSortAscending, // Function { String } columnName
        onSortClear, // Function { String } columnName
        onSortDescending, // Function { String } columnName
        pageSize, // Number
        rows, // [ Object ]
        selectedRowIndexes, // [ Number ]
        showFilterRow, // Boolean
        sortDirectives, // [ { columnName: String, direction: 'asc'|'desc'|null } ]
        testID, // String
        totalRows, // Number
    } = props;

    console.log('DATA TABLE RXG: RENDER: ', props); // eslint-disable-line
    console.log('DATA TABLE RXG: RENDER PROPS STRING: ', JSON.stringify(props)); // eslint-disable-line
    console.log('DATA TABLE RXG: SORT DIRECTIVES: ', sortDirectives); // eslint-disable-line
    console.log('DATA TABLE RXG: FILTER DIRECTIVES: ', filterDirectives); // eslint-disable-line
    console.log('DATA TABLE RXG: FILTER OPERATIONS: ', filterOperations); // eslint-disable-line
    console.log('DATA TABLE RXG: FILTER OPERATIONS STRING: ', JSON.stringify(filterOperations)); // eslint-disable-line

    const [
        selection,
        setSelection,
    ] = useState(selectedRowIndexes);

    useEffect(() => {
        setSelection(selectedRowIndexes);
    }, [ selectedRowIndexes ]);

    const [
        filters,
        setFilters,
    ] = useState(filterDirectives);

    const styles = getStyles(contextStyles);
    const columnOrder = columns.map((c) => (c.name));
    const columnWidths = columns.map((c) => {
        return {
            columnName: c.name,
            width: c.width || defaultColumnWidth,
        };
    });
    const fixedLeftColumns = [];
    const fixedColumnsMap = {};
    fixedColumnIndexes.forEach((idx) => {
        const columnName = columns[idx].name;
        fixedLeftColumns.push(columnName);
        fixedColumnsMap[columnName] = idx;
    });

    return (
        <div
            className="c-data-table__container"
            style={ styles.container }
            data-test-id={ `${testID}__data-table__container` }>
            { loading && <ActivityIndicator centered /> }
            <Grid
                columns={ columns }
                rows={ rows }>

                { /* Include Custom Filtering Drop Down Operations */ }
                { /* filterOperations.map((filterOperation) => {
                    const {
                        columnName,
                        operators,
                    } = filterOperation;
                    return (
                        <DataTypeProvider
                            key={ columnName }
                            for={ [ columnName ] }
                            availableFilterOperations={ operators } />
                    );
                }) */ }

                { /* Include Infinite Scrolling Functionality and Handling */ }
                { infiniteScrolling &&
                    <VirtualTableState
                        infiniteScrolling={ infiniteScrolling }
                        loading={ loading }
                        pageSize={ pageSize }
                        totalRowCount={ totalRows }
                        skip={ 0 }
                        getRows={ onLoadMoreData } />
                }

                { /* Include Row Selection Handling */ }
                <SelectionState
                    selection={ selection }
                    onSelectionChange={ (...args) => {
                        setSelection(...args);
                        if (onRowSelectionChange) {
                            onRowSelectionChange(...args);
                        }
                    } } />

                { /* Include Filtering Functionality and Handling */ }
                { showFilterRow &&
                    <FilteringState
                        filters={ filters }
                        onFiltersChange={ (...args) => {
                            setFilters(...args);
                            if (onFilterChange) {
                                onFilterChange(...args);
                            }
                        } } />
                }

                { /* Include Drag and Drop Functionality and Handling */ }
                <DragDropProvider />

                { /* Include Table */ }
                <VirtualTable
                    cellComponent={ cellComponent } />

                { /* Include Column Reordering Functionality and Handling */ }
                { /* Controlled With Internal Grid State */ }
                <TableColumnReordering
                    defaultOrder={ columnOrder }
                    onOrderChange={ onColumnsReordered } />

                { /* Include Column Resizing Functionality and Handling */ }
                { /* Controlled With Internal Grid State */ }
                <TableColumnResizing
                    defaultColumnWidths={ columnWidths }
                    onColumnWidthsChange={ onColumnsResized } />

                { /* Include Column Header Functionality and Handling */ }
                <TableHeaderRow
                    contentComponent={ (p) => (
                        <HeaderCellContent
                            { ...p }
                            onFilterToggle={ onFilterToggle }
                            onColumnFrozen={ onColumnFrozen }
                            fixedColumns={ fixedColumnsMap }
                            showFilterRow={ showFilterRow } />
                    ) } />

                { /* Include Filter Row when enabled */ }
                { showFilterRow &&
                    <TableFilterRow
                        cellComponent={ (p) => (
                            <FilterCell
                                { ...p }
                                onFilterCancel={ onFilterCancel }
                                onFilterReset={ onFilterReset }
                                onFilterSubmit={ onFilterSubmit }
                                onSortAscending={ onSortAscending }
                                onSortClear={ onSortClear }
                                onSortDescending={ onSortDescending }
                                sortDirectives={ sortDirectives } />
                        ) }
                        rowHeight={ 145 }
                        showFilterSelector />
                }

                { /* Include Row Selection Functionality */ }
                <TableSelection
                    selectByRowClick
                    highlightRow
                    showSelectionColumn={ false } />

                { /* Include Fixed/Frozen/Pinned columns */ }
                <TableFixedColumns
                    leftColumns={ fixedLeftColumns } />
            </Grid>
        </div>
    );
};

DataTable.propTypes = {
    /** Component used to generate the table cell and its contents */
    cellComponent: PropTypes.oneOfType([
        PropTypes.func,
        PropTypes.object,
    ]),

    /** An array containing column data */
    columns: PropTypes.arrayOf(PropTypes.shape({
        /** The name identifier of the column */
        name: PropTypes.string,

        /** The column header display text */
        title: PropTypes.string,

        /** Overrides the default table column width in pixels or CSS-accepted units (auto, px, %, em, rem, vm, vh, vmin, vmax) */
        width: PropTypes.oneOfType([
            PropTypes.number,
            PropTypes.string,
        ]),
    })).isRequired,

    /** Styles for this component */
    contextStyles: PropTypes.shape({
        /** Styles for the container surrounding the table */
        container: PropTypes.object,
    }),

    /** Default (px) width applied to all columns */
    defaultColumnWidth: PropTypes.number,

    /** Filter entry directives */
    filterDirectives: PropTypes.arrayOf(PropTypes.shape({
        /** The name identifier of the filter column */
        columnName: PropTypes.string,

        /** The filter operation */
        operation: PropTypes.string,

        /** The filter value */
        value: PropTypes.oneOfType([
            PropTypes.string,
            PropTypes.number,
            PropTypes.bool,
        ]),
    })),

    /** Column filter operators */
    filterOperations: PropTypes.arrayOf(PropTypes.shape({
        /** The name identifier of the filter column */
        columnName: PropTypes.string,

        /** The filter operation */
        operators: PropTypes.arrayOf(PropTypes.string),
    })),

    /**
     * Indexes of columns to freeze and affix to the left of the table
     * NOTE: non-contiguous column indexes do NOT reorder columns and dock together as a group.
     * If fixed grouped non-contiguous columns are desired then the columns prop must be reordered with
     * fixed columns at the beginning of the columns array in the desired fixed order.
     */
    fixedColumnIndexes: PropTypes.arrayOf(PropTypes.number),

    /** Enables a scrolling mode in which rows are loaded in sequence */
    infiniteScrolling: PropTypes.bool,

    /** Shows the table loading state when enabled */
    loading: PropTypes.bool,

    /**
     * Called on column freeze toggle component click
     * @param {String} columnName - the name of the column with the freeze toggle
     * @returns {void}
     */
    onColumnFrozen: PropTypes.func,

    /**
     * Called on column reorder
     * @param {Array.<String>} columnsNamesInOrder
     * @returns {void}
     */
    onColumnsReordered: PropTypes.func,

    /**
     * Called on column resize
     * @param {Array.<{ columnName: String, width: Number }>} columnsWithWidths
     * @returns {void}
     */
    onColumnsResized: PropTypes.func,

    /**
     * Called on column filter cancel component click
     * @param {String} columnName - the name identifier of the column
     * @returns {void}
     */
    onFilterCancel: PropTypes.func,

    /**
     * @typedef {Object} Filter
     * @property {String} columnName
     * @property {FilterOperations} operation
     * @property {String} value
     *
     * Called on filter change
     * @param {Array.<Filter>} filters
     * @returns {void}
     */
    onFilterChange: PropTypes.func,

    /**
     * Called on column filters reset component click
     * @param {String} columnName - the name identifier of the column
     * @returns {void}
     */
    onFilterReset: PropTypes.func,

    /**
     * Called on column filters submit component click
     * @param {String} columnName - the name identifier of the column
     * @returns {void}
     */
    onFilterSubmit: PropTypes.func,

    /**
     * Called on column filter toggle component click
     * @param {String} columnName - the name identifier of the column
     * @returns {void}
     */
    onFilterToggle: PropTypes.func,

    /**
     * Called when the viewport row threshold reaches a certain point when scrolling
     * @param {Number} skip
     * @param {Number} take
     * @returns {void}
     */
    onLoadMoreData: PropTypes.func,

    /**
     * Called on row click
     * @param {Array.<Number|String>} selectedRowIndexes
     * @returns {void}
     */
    onRowSelectionChange: PropTypes.func,

    /**
     * Called on column sort ascending toggle component click
     * @param {String} columnName - the name of the column to be sorted
     * @returns {void}
     */
    onSortAscending: PropTypes.func,

    /**
     * Called on column sort clear toggle component click
     * @param {String} columnName - the name of the column to clear sorting
     * @returns {void}
     */
    onSortClear: PropTypes.func,

    /**
     * Called on column sort descending toggle component click
     * @param {String} columnName - the name of the column to be sorted
     * @returns {void}
     */
    onSortDescending: PropTypes.func,

    /** Specifies the count of rows in the current virtual chunk */
    pageSize: PropTypes.number,

    /** An array containing custom data */
    rows: PropTypes.array.isRequired,

    /** Indexes of rows that show as selected in the UI */
    selectedRowIndexes: PropTypes.arrayOf(PropTypes.number),

    /** Shows the table header filter row when enabled */
    showFilterRow: PropTypes.bool,

    /** Column sorting directives */
    sortDirectives: PropTypes.arrayOf(PropTypes.shape({
        /** The name identifier of the column */
        columnName: PropTypes.string,

        /** The sorting direction reflected in the UI */
        direction: PropTypes.oneOf([
            'asc',
            'desc',
            null,
        ]),
    })),

    /** ID string used for testing */
    testID: PropTypes.string,

    /** The total number of rows to render including virtual rows when infinite scrolling is enabled */
    totalRows: PropTypes.number.isRequired,
};

DataTable.defaultProps = {
    cellComponent: DataTableCell,
    contextStyles: {},
    defaultColumnWidth: DEFAULT_COLUMN_WIDTH,
    filterDirectives: [],
    filterOperations: [],
    fixedColumnIndexes: [],
    infiniteScrolling: false,
    selectedRowIndexes: [],
    sortDirectives: [],
    testID: 'data-table-root',
};

export {
    DataTableCell,
    FILTER_OPERATIONS,
};

export default DataTable;
